@@ -3,13 +3,16 @@
 #include <math.h>
 
 #include "access/generic_xlog.h"
+#include "catalog/pg_type.h"
+#include "catalog/pg_type_d.h"
+#include "fmgr.h"
 #include "hnsw.h"
 #include "lib/pairingheap.h"
+#include "sparsevec.h"
 #include "storage/bufmgr.h"
 #include "utils/datum.h"
 #include "utils/memdebug.h"
 #include "utils/rel.h"
-#include "vector.h"
 
 #if PG_VERSION_NUM >= 130000
 #include "common/hashfn.h"
@@ -150,34 +153,21 @@ HnswOptionalProcInfo(Relation index, uint16 procnum)
 }
 
 /*
- * Divide by the norm
- *
- * Returns false if value should not be indexed
- *
- * The caller needs to free the pointer stored in value
- * if it's different than the original value
+ * Normalize value
+ */
+Datum
+HnswNormValue(const HnswTypeInfo * typeInfo, Oid collation, Datum value)
+{
+	return DirectFunctionCall1Coll(typeInfo->normalize, collation, value);
+}
+
+/*
+ * Check if non-zero norm
  */
 bool
-HnswNormValue(FmgrInfo *procinfo, Oid collation, Datum *value, Vector * result)
+HnswCheckNorm(FmgrInfo *procinfo, Oid collation, Datum value)
 {
-	double		norm = DatumGetFloat8(FunctionCall1Coll(procinfo, collation, *value));
-
-	if (norm > 0)
-	{
-		Vector	   *v = DatumGetVector(*value);
-
-		if (result == NULL)
-			result = InitVector(v->dim);
-
-		for (int i = 0; i < v->dim; i++)
-			result->x[i] = v->x[i] / norm;
-
-		*value = PointerGetDatum(result);
-
-		return true;
-	}
-
-	return false;
+	return DatumGetFloat8(FunctionCall1Coll(procinfo, collation, value)) > 0;
 }
 
 /*
@@ -310,6 +300,9 @@ HnswGetMetaPageInfo(Relation index, int *m, HnswElement * entryPoint)
 	page = BufferGetPage(buf);
 	metap = HnswPageGetMeta(page);
 
+	if (unlikely(metap->magicNumber != HNSW_MAGIC_NUMBER))
+		elog(ERROR, "hnsw index is not valid");
+
 	if (m != NULL)
 		*m = metap->m;
 
@@ -555,7 +548,7 @@ HnswLoadElementFromTuple(HnswElement element, HnswElementTuple etup, bool loadHe
  * Load an element and optionally get its distance from q
  */
 void
-HnswLoadElement(HnswElement element, float *distance, Datum *q, Relation index, FmgrInfo *procinfo, Oid collation, bool loadVec)
+HnswLoadElement(HnswElement element, float *distance, Datum *q, Relation index, FmgrInfo *procinfo, Oid collation, bool loadVec, float *maxDistance)
 {
 	Buffer		buf;
 	Page		page;
@@ -570,12 +563,18 @@ HnswLoadElement(HnswElement element, float *distance, Datum *q, Relation index,
 
 	Assert(HnswIsElementTuple(etup));
 
-	/* Load element */
-	HnswLoadElementFromTuple(element, etup, true, loadVec);
-
 	/* Calculate distance */
 	if (distance != NULL)
-		*distance = (float) DatumGetFloat8(FunctionCall2Coll(procinfo, collation, *q, PointerGetDatum(&etup->data)));
+	{
+		if (DatumGetPointer(*q) == NULL)
+			*distance = 0;
+		else
+			*distance = (float) DatumGetFloat8(FunctionCall2Coll(procinfo, collation, *q, PointerGetDatum(&etup->data)));
+	}
+
+	/* Load element */
+	if (distance == NULL || maxDistance == NULL || *distance < *maxDistance)
+		HnswLoadElementFromTuple(element, etup, true, loadVec);
 
 	UnlockReleaseBuffer(buf);
 }
@@ -604,7 +603,7 @@ HnswEntryCandidate(char *base, HnswElement entryPoint, Datum q, Relation index,
 	if (index == NULL)
 		hc->distance = GetCandidateDistance(base, hc, q, procinfo, collation);
 	else
-		HnswLoadElement(entryPoint, &hc->distance, &q, index, procinfo, collation, loadVec);
+		HnswLoadElement(entryPoint, &hc->distance, &q, index, procinfo, collation, loadVec, NULL);
 	return hc;
 }
 
@@ -800,25 +799,27 @@ HnswSearchLayer(char *base, Datum q, List *ep, int ef, int lc, Relation index, F
 			{
 				float		eDistance;
 				HnswElement eElement = HnswPtrAccess(base, e->element);
+				bool		alwaysAdd = wlen < ef;
 
 				f = ((HnswPairingHeapNode *) pairingheap_first(W))->inner;
 
 				if (index == NULL)
 					eDistance = GetCandidateDistance(base, e, q, procinfo, collation);
 				else
-					HnswLoadElement(eElement, &eDistance, &q, index, procinfo, collation, inserting);
+					HnswLoadElement(eElement, &eDistance, &q, index, procinfo, collation, inserting, alwaysAdd ? NULL : &f->distance);
 
-				Assert(!eElement->deleted);
-
-				/* Make robust to issues */
-				if (eElement->level < lc)
-					continue;
-
-				if (eDistance < f->distance || wlen < ef)
+				if (eDistance < f->distance || alwaysAdd)
 				{
-					/* Copy e */
-					HnswCandidate *ec = palloc(sizeof(HnswCandidate));
+					HnswCandidate *ec;
 
+					Assert(!eElement->deleted);
+
+					/* Make robust to issues */
+					if (eElement->level < lc)
+						continue;
+
+					/* Copy e */
+					ec = palloc(sizeof(HnswCandidate));
 					HnswPtrStore(base, ec->element, eElement);
 					ec->distance = eDistance;
 
@@ -860,12 +861,15 @@ HnswSearchLayer(char *base, Datum q, List *ep, int ef, int lc, Relation index, F
 static int
 #if PG_VERSION_NUM >= 130000
 CompareCandidateDistances(const ListCell *a, const ListCell *b)
+{
+	HnswCandidate *hca = lfirst(a);
+	HnswCandidate *hcb = lfirst(b);
 #else
 CompareCandidateDistances(const void *a, const void *b)
-#endif
 {
-	HnswCandidate *hca = lfirst((ListCell *) a);
-	HnswCandidate *hcb = lfirst((ListCell *) b);
+	HnswCandidate *hca = lfirst(*(ListCell **) a);
+	HnswCandidate *hcb = lfirst(*(ListCell **) b);
+#endif
 
 	if (hca->distance < hcb->distance)
 		return 1;
@@ -888,12 +892,15 @@ CompareCandidateDistances(const void *a, const void *b)
 static int
 #if PG_VERSION_NUM >= 130000
 CompareCandidateDistancesOffset(const ListCell *a, const ListCell *b)
+{
+	HnswCandidate *hca = lfirst(a);
+	HnswCandidate *hcb = lfirst(b);
 #else
 CompareCandidateDistancesOffset(const void *a, const void *b)
-#endif
 {
-	HnswCandidate *hca = lfirst((ListCell *) a);
-	HnswCandidate *hcb = lfirst((ListCell *) b);
+	HnswCandidate *hca = lfirst(*(ListCell **) a);
+	HnswCandidate *hcb = lfirst(*(ListCell **) b);
+#endif
 
 	if (hca->distance < hcb->distance)
 		return 1;
@@ -952,7 +959,9 @@ SelectNeighbors(char *base, List *c, int lm, int lc, FmgrInfo *procinfo, Oid col
 {
 	List	   *r = NIL;
 	List	   *w = list_copy(c);
-	pairingheap *wd;
+	HnswCandidate **wd;
+	int			wdlen = 0;
+	int			wdoff = 0;
 	HnswNeighborArray *neighbors = HnswGetNeighbors(base, e2, lc);
 	bool		mustCalculate = !neighbors->closerSet;
 	List	   *added = NIL;
@@ -961,7 +970,7 @@ SelectNeighbors(char *base, List *c, int lm, int lc, FmgrInfo *procinfo, Oid col
 	if (list_length(w) <= lm)
 		return w;
 
-	wd = pairingheap_allocate(CompareNearestCandidates, NULL);
+	wd = palloc(sizeof(HnswCandidate *) * list_length(w));
 
 	/* Ensure order of candidates is deterministic for closer caching */
 	if (sortCandidates)
@@ -1027,21 +1036,21 @@ SelectNeighbors(char *base, List *c, int lm, int lc, FmgrInfo *procinfo, Oid col
 		if (e->closer)
 			r = lappend(r, e);
 		else
-			pairingheap_add(wd, &(CreatePairingHeapNode(e)->ph_node));
+			wd[wdlen++] = e;
 	}
 
 	/* Cached value can only be used in future if sorted deterministically */
 	neighbors->closerSet = sortCandidates;
 
 	/* Keep pruned connections */
-	while (!pairingheap_is_empty(wd) && list_length(r) < lm)
-		r = lappend(r, ((HnswPairingHeapNode *) pairingheap_remove_first(wd))->inner);
+	while (wdoff < wdlen && list_length(r) < lm)
+		r = lappend(r, wd[wdoff++]);
 
 	/* Return pruned for update connections */
 	if (pruned != NULL)
 	{
-		if (!pairingheap_is_empty(wd))
-			*pruned = ((HnswPairingHeapNode *) pairingheap_first(wd))->inner;
+		if (wdoff < wdlen)
+			*pruned = wd[wdoff];
 		else
 			*pruned = linitial(w);
 	}
@@ -1099,7 +1108,7 @@ HnswUpdateConnection(char *base, HnswElement element, HnswCandidate * hc, int lm
 				HnswElement hc3Element = HnswPtrAccess(base, hc3->element);
 
 				if (HnswPtrIsNull(base, hc3Element->value))
-					HnswLoadElement(hc3Element, &hc3->distance, &q, index, procinfo, collation, true);
+					HnswLoadElement(hc3Element, &hc3->distance, &q, index, procinfo, collation, true, NULL);
 				else
 					hc3->distance = GetCandidateDistance(base, hc3, q, procinfo, collation);
 
@@ -1260,3 +1269,77 @@ HnswFindElementNeighbors(char *base, HnswElement element, HnswElement entryPoint
 		ep = w;
 	}
 }
+
+PGDLLEXPORT Datum l2_normalize(PG_FUNCTION_ARGS);
+PGDLLEXPORT Datum halfvec_l2_normalize(PG_FUNCTION_ARGS);
+PGDLLEXPORT Datum sparsevec_l2_normalize(PG_FUNCTION_ARGS);
+
+static void
+SparsevecCheckValue(Pointer v)
+{
+	SparseVector *vec = (SparseVector *) v;
+
+	if (vec->nnz > HNSW_MAX_NNZ)
+		elog(ERROR, "sparsevec cannot have more than %d non-zero elements for hnsw index", HNSW_MAX_NNZ);
+}
+
+/*
+ * Get type info
+ */
+const		HnswTypeInfo *
+HnswGetTypeInfo(Relation index)
+{
+	FmgrInfo   *procinfo = HnswOptionalProcInfo(index, HNSW_TYPE_INFO_PROC);
+
+	if (procinfo == NULL)
+	{
+		static const HnswTypeInfo typeInfo = {
+			.maxDimensions = HNSW_MAX_DIM,
+			.normalize = l2_normalize,
+			.checkValue = NULL
+		};
+
+		return (&typeInfo);
+	}
+	else
+		return (const HnswTypeInfo *) DatumGetPointer(FunctionCall0Coll(procinfo, InvalidOid));
+}
+
+PGDLLEXPORT PG_FUNCTION_INFO_V1(hnsw_halfvec_support);
+Datum
+hnsw_halfvec_support(PG_FUNCTION_ARGS)
+{
+	static const HnswTypeInfo typeInfo = {
+		.maxDimensions = HNSW_MAX_DIM * 2,
+		.normalize = halfvec_l2_normalize,
+		.checkValue = NULL
+	};
+
+	PG_RETURN_POINTER(&typeInfo);
+};
+
+PGDLLEXPORT PG_FUNCTION_INFO_V1(hnsw_bit_support);
+Datum
+hnsw_bit_support(PG_FUNCTION_ARGS)
+{
+	static const HnswTypeInfo typeInfo = {
+		.maxDimensions = HNSW_MAX_DIM * 32,
+		.normalize = NULL,
+		.checkValue = NULL
+	};
+
+	PG_RETURN_POINTER(&typeInfo);
+};
+
+PGDLLEXPORT PG_FUNCTION_INFO_V1(hnsw_sparsevec_support);
+Datum
+hnsw_sparsevec_support(PG_FUNCTION_ARGS)
+{
+	static const HnswTypeInfo typeInfo = {
+		.maxDimensions = SPARSEVEC_MAX_DIM,
+		.normalize = sparsevec_l2_normalize,
+		.checkValue = SparsevecCheckValue
+	};
+
+	PG_RETURN_POINTER(&typeInfo);
+};
