@@ -12,7 +12,7 @@ typedef struct Vector
 {
 	int32		vl_len_;		/* varlena header (do not touch directly!) */
 	int16		dim;			/* number of dimensions */
-	int16		unused;
+	int16		unused;			/* reserved for future use, always zero */
 	float		x[FLEXIBLE_ARRAY_MEMBER];
 }			Vector;
 
